import time
import json
import copy
import base64
from urllib import parse

import grpc
import numpy as np
from grpc._cython import cygrpc

from ..grpc_gen import milvus_pb2_grpc
from ..grpc_gen import milvus_pb2 as milvus_types
from ..grpc_gen import common_pb2

from .abstract import CollectionSchema, ChunkedQueryResult, MutationResult
from .check import (
    is_legal_host,
    is_legal_port,
    check_pass_param,
    check_index_params,
)
from .prepare import Prepare
from .types import (
    Status,
    IndexState,
    LoadState,
    DataType,
    CompactionState,
    State,
    CompactionPlans,
    Plan,
    get_consistency_level,
    Replica, Shard, Group,
    GrantInfo, UserInfo, RoleInfo,
    BulkInsertState,
    ResourceGroupInfo,
)

from .utils import (
    check_invalid_binary_vector,
    len_of
)

from ..settings import DefaultConfig as config
from .configs import DefaultConfigs
from . import ts_utils
from . import interceptor

from .asynch import (
    SearchFuture,
    MutationFuture,
    CreateIndexFuture,
    FlushFuture,
    LoadPartitionsFuture,
    ChunkedSearchFuture
)

from ..exceptions import (
    ExceptionsMessage,
    ParamError,
    DescribeCollectionException,
    MilvusException,
    AmbiguousIndexName,
)

from ..decorators import retry_on_rpc_failure


class GrpcHandler:
    def __init__(self, uri=config.GRPC_URI, host="", port="", channel=None, **kwargs):
        self._stub = None
        self._channel = channel

        addr = kwargs.get("address")
        self._address = addr if addr is not None else self.__get_address(uri, host, port)
        self._log_level = None
        self._request_id = None
        self._set_authorization(**kwargs)
        self._setup_grpc_channel()

    def __get_address(self, uri: str, host: str, port: str) -> str:
        if host != "" and port != "" and is_legal_host(host) and is_legal_port(port):
            return f"{host}:{port}"

        try:
            parsed_uri = parse.urlparse(uri)
        except (Exception) as e:
            raise ParamError(message=f"Illegal uri: [{uri}], {e}") from e
        return parsed_uri.netloc

    def _set_authorization(self, **kwargs):
        secure = kwargs.get("secure", False)
        if not isinstance(secure, bool):
            raise ParamError(message="secure must be bool type")
        self._secure = secure
        self._client_pem_path = kwargs.get("client_pem_path", "")
        self._client_key_path = kwargs.get("client_key_path", "")
        self._ca_pem_path = kwargs.get("ca_pem_path", "")
        self._server_pem_path = kwargs.get("server_pem_path", "")
        self._server_name = kwargs.get("server_name", "")

        self._authorization_interceptor = None
        self._setup_authorization_interceptor(kwargs.get("user", None), kwargs.get("password", None))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _wait_for_channel_ready(self, timeout=10):
        if self._channel is not None:
            try:
                grpc.channel_ready_future(self._channel).result(timeout=timeout)
                return
            except grpc.FutureTimeoutError as e:
                raise MilvusException(Status.CONNECT_FAILED,
                                      f'Fail connecting to server on {self._address}. Timeout') from e

        raise MilvusException(Status.CONNECT_FAILED, 'No channel in handler, please setup grpc channel first')

    def close(self):
        self._channel.close()

    def _setup_authorization_interceptor(self, user, password):
        if user and password:
            authorization = base64.b64encode(f"{user}:{password}".encode('utf-8'))
            key = "authorization"
            self._authorization_interceptor = interceptor.header_adder_interceptor(key, authorization)

    def _setup_grpc_channel(self):
        """ Create a ddl grpc channel """
        if self._channel is None:
            opts = [(cygrpc.ChannelArgKey.max_send_message_length, -1),
                    (cygrpc.ChannelArgKey.max_receive_message_length, -1),
                    ('grpc.enable_retries', 1),
                    ('grpc.keepalive_time_ms', 55000),
                    ]
            if not self._secure:
                self._channel = grpc.insecure_channel(
                    self._address,
                    options=opts,
                )
            else:
                if self._client_pem_path != "" and self._client_key_path != "" and self._ca_pem_path != "" \
                        and self._server_name != "":
                    opts.append(('grpc.ssl_target_name_override', self._server_name, ),)
                    with open(self._client_pem_path, 'rb') as f:
                        certificate_chain = f.read()
                    with open(self._client_key_path, 'rb') as f:
                        private_key = f.read()
                    with open(self._ca_pem_path, 'rb') as f:
                        root_certificates = f.read()
                    creds = grpc.ssl_channel_credentials(root_certificates, private_key, certificate_chain)
                elif self._server_pem_path != "" and self._server_name != "":
                    opts.append(('grpc.ssl_target_name_override', self._server_name,), )
                    with open(self._server_pem_path, 'rb') as f:
                        server_pem = f.read()
                    creds = grpc.ssl_channel_credentials(root_certificates=server_pem)
                else:
                    creds = grpc.ssl_channel_credentials(root_certificates=None, private_key=None,
                                                         certificate_chain=None)
                self._channel = grpc.secure_channel(
                    self._address,
                    creds,
                    options=opts
                )
        # avoid to add duplicate headers.
        self._final_channel = self._channel
        if self._authorization_interceptor:
            self._final_channel = grpc.intercept_channel(self._final_channel, self._authorization_interceptor)
        if self._log_level:
            log_level_interceptor = interceptor.header_adder_interceptor("log_level", self._log_level)
            self._final_channel = grpc.intercept_channel(self._final_channel, log_level_interceptor)
            self._log_level = None
        if self._request_id:
            request_id_interceptor = interceptor.header_adder_interceptor("client_request_id", self._request_id)
            self._final_channel = grpc.intercept_channel(self._final_channel, request_id_interceptor)
            self._request_id = None
        self._stub = milvus_pb2_grpc.MilvusServiceStub(self._final_channel)

    def set_onetime_loglevel(self, log_level):
        self._log_level = log_level
        self._setup_grpc_channel()

    def set_onetime_request_id(self, req_id):
        self._request_id = req_id
        self._setup_grpc_channel()

    @property
    def server_address(self):
        """ Server network address """
        return self._address

    def reset_password(self, user, old_password, new_password, timeout=None):
        """
        reset password and then setup the grpc channel.
        """
        self.update_password(user, old_password, new_password, timeout=timeout)
        self._setup_authorization_interceptor(user, new_password)
        self._setup_grpc_channel()

    @retry_on_rpc_failure()
    def create_collection(self, collection_name, fields, shards_num=2, timeout=None, **kwargs):
        request = Prepare.create_collection_request(collection_name, fields, shards_num=shards_num, **kwargs)

        rf = self._stub.CreateCollection.future(request, timeout=timeout)
        if kwargs.get("_async", False):
            return rf
        status = rf.result()
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure(retry_on_deadline=False)
    def drop_collection(self, collection_name, timeout=None):
        check_pass_param(collection_name=collection_name)
        request = Prepare.drop_collection_request(collection_name)

        rf = self._stub.DropCollection.future(request, timeout=timeout)
        status = rf.result()
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def alter_collection(self, collection_name, properties, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name, properties=properties)
        request = Prepare.alter_collection_request(collection_name, properties)
        rf = self._stub.AlterCollection.future(request, timeout=timeout)
        status = rf.result()
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def has_collection(self, collection_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.describe_collection_request(collection_name)
        rf = self._stub.DescribeCollection.future(request, timeout=timeout)

        reply = rf.result()
        if reply.status.error_code == common_pb2.Success:
            return True

        # TODO: Workaround for unreasonable describe collection results and error_code
        if reply.status.error_code == common_pb2.UnexpectedError and "can\'t find collection" in reply.status.reason:
            return False

        raise MilvusException(reply.status.error_code, reply.status.reason)

    @retry_on_rpc_failure()
    def describe_collection(self, collection_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.describe_collection_request(collection_name)
        rf = self._stub.DescribeCollection.future(request, timeout=timeout)
        response = rf.result()
        status = response.status

        if status.error_code == 0:
            return CollectionSchema(raw=response).dict()

        raise DescribeCollectionException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def list_collections(self, timeout=None):
        request = Prepare.show_collections_request()
        rf = self._stub.ShowCollections.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if response.status.error_code == 0:
            return list(response.collection_names)

        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def rename_collections(self, old_name=None, new_name=None, timeout=None):
        check_pass_param(collection_name=new_name)
        check_pass_param(collection_name=old_name)
        request = Prepare().rename_collections_request(old_name, new_name)
        rf = self._stub.RenameCollection.future(request, timeout=timeout)
        response = rf.result()

        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def create_partition(self, collection_name, partition_name, timeout=None):
        check_pass_param(collection_name=collection_name, partition_name=partition_name)
        request = Prepare.create_partition_request(collection_name, partition_name)
        rf = self._stub.CreatePartition.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def drop_partition(self, collection_name, partition_name, timeout=None):
        check_pass_param(collection_name=collection_name, partition_name=partition_name)
        request = Prepare.drop_partition_request(collection_name, partition_name)

        rf = self._stub.DropPartition.future(request, timeout=timeout)
        response = rf.result()

        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def has_partition(self, collection_name, partition_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name, partition_name=partition_name)
        request = Prepare.has_partition_request(collection_name, partition_name)
        rf = self._stub.HasPartition.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            return response.value

        raise MilvusException(status.error_code, status.reason)

    # TODO: this is not inuse
    @retry_on_rpc_failure()
    def get_partition_info(self, collection_name, partition_name, timeout=None):
        request = Prepare.partition_stats_request(collection_name, partition_name)
        rf = self._stub.DescribePartition.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            statistics = response.statistics
            info_dict = {}
            for kv in statistics:
                info_dict[kv.key] = kv.value
            return info_dict
        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def list_partitions(self, collection_name, timeout=None):
        check_pass_param(collection_name=collection_name)
        request = Prepare.show_partitions_request(collection_name)

        rf = self._stub.ShowPartitions.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            return list(response.partition_names)

        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def get_partition_stats(self, collection_name, partition_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        req = Prepare.get_partition_stats_request(collection_name, partition_name)
        future = self._stub.GetPartitionStatistics.future(req, timeout=timeout)
        response = future.result()
        status = response.status
        if status.error_code == 0:
            return response.stats

        raise MilvusException(status.error_code, status.reason)

    def _prepare_batch_insert_request(self, collection_name, entities, partition_name=None, timeout=None, **kwargs):
        insert_param = kwargs.get('insert_param', None)

        if insert_param and not isinstance(insert_param, milvus_types.RowBatch):
            raise ParamError(message="The value of key 'insert_param' is invalid")
        if not isinstance(entities, list):
            raise ParamError(message="None entities, please provide valid entities.")

        collection_schema = kwargs.get("schema", None)
        if not collection_schema:
            collection_schema = self.describe_collection(collection_name, timeout=timeout, **kwargs)

        fields_info = collection_schema["fields"]

        request = insert_param if insert_param \
            else Prepare.batch_insert_param(collection_name, entities, partition_name, fields_info)

        return request

    @retry_on_rpc_failure()
    def batch_insert(self, collection_name, entities, partition_name=None, timeout=None, **kwargs):
        if not check_invalid_binary_vector(entities):
            raise ParamError(message="Invalid binary vector data exists")

        try:
            request = self._prepare_batch_insert_request(collection_name, entities, partition_name, timeout, **kwargs)
            rf = self._stub.Insert.future(request, timeout=timeout)
            if kwargs.get("_async", False) is True:
                cb = kwargs.get("_callback", None)
                f = MutationFuture(rf, cb, timeout=timeout, **kwargs)
                f.add_callback(ts_utils.update_ts_on_mutation(collection_name))
                return f

            response = rf.result()
            if response.status.error_code == 0:
                m = MutationResult(response)
                ts_utils.update_collection_ts(collection_name, m.timestamp)
                return m

            raise MilvusException(response.status.error_code, response.status.reason)
        except Exception as err:
            if kwargs.get("_async", False):
                return MutationFuture(None, None, err)
            raise err

    @retry_on_rpc_failure()
    def delete(self, collection_name, expression, partition_name=None, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        try:
            req = Prepare.delete_request(collection_name, partition_name, expression)
            future = self._stub.Delete.future(req, timeout=timeout)

            if kwargs.get("_async", False):
                cb = kwargs.get("_callback", None)
                f = MutationFuture(future, cb, timeout=timeout, **kwargs)
                f.add_callback(ts_utils.update_ts_on_mutation(collection_name))
                return f

            response = future.result()
            if response.status.error_code == 0:
                m = MutationResult(response)
                ts_utils.update_collection_ts(collection_name, m.timestamp)
                return m

            raise MilvusException(response.status.error_code, response.status.reason)
        except Exception as err:
            if kwargs.get("_async", False):
                return MutationFuture(None, None, err)
            raise err

    def _execute_search_requests(self, requests, timeout=None, **kwargs):
        auto_id = kwargs.get("auto_id", True)

        try:
            if kwargs.get("_async", False):
                futures = []
                for request in requests:
                    ft = self._stub.Search.future(request, timeout=timeout)
                    futures.append(ft)
                func = kwargs.get("_callback", None)
                return ChunkedSearchFuture(futures, func, auto_id)

            raws = []
            for request in requests:
                response = self._stub.Search(request, timeout=timeout)

                if response.status.error_code != 0:
                    raise MilvusException(response.status.error_code, response.status.reason)

                raws.append(response)
            round_decimal = kwargs.get("round_decimal", -1)
            return ChunkedQueryResult(raws, auto_id, round_decimal)

        except Exception as pre_err:
            if kwargs.get("_async", False):
                return SearchFuture(None, None, True, pre_err)
            raise pre_err

    @retry_on_rpc_failure(retry_on_deadline=False)
    def search(self, collection_name, data, anns_field, param, limit,
               expression=None, partition_names=None, output_fields=None,
               round_decimal=-1, timeout=None, schema=None, **kwargs):
        check_pass_param(
            limit=limit,
            round_decimal=round_decimal,
            anns_field=anns_field,
            search_data=data,
            partition_name_array=partition_names,
            output_fields=output_fields,
            travel_timestamp=kwargs.get("travel_timestamp", 0),
            guarantee_timestamp=kwargs.get("guarantee_timestamp", 0)
        )

        if schema is None:
            schema = self.describe_collection(collection_name, timeout=timeout, **kwargs)

        consistency_level = schema["consistency_level"]
        # overwrite the consistency level defined when user created the collection
        consistency_level = get_consistency_level(kwargs.get("consistency_level", consistency_level))

        ts_utils.construct_guarantee_ts(consistency_level, collection_name, kwargs)

        requests = Prepare.search_requests_with_expr(collection_name, data, anns_field, param, limit, schema,
                                                     expression, partition_names, output_fields, round_decimal,
                                                     **kwargs)

        auto_id = schema["auto_id"]
        return self._execute_search_requests(requests, timeout, round_decimal=round_decimal, auto_id=auto_id, **kwargs)

    @retry_on_rpc_failure()
    def get_query_segment_info(self, collection_name, timeout=30, **kwargs):
        req = Prepare.get_query_segment_info_request(collection_name)
        future = self._stub.GetQuerySegmentInfo.future(req, timeout=timeout)
        response = future.result()
        status = response.status
        if status.error_code == 0:
            return response.infos  # todo: A wrapper class of QuerySegmentInfo
        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def create_alias(self, collection_name, alias, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.create_alias_request(collection_name, alias)
        rf = self._stub.CreateAlias.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def drop_alias(self, alias, timeout=None, **kwargs):
        request = Prepare.drop_alias_request(alias)
        rf = self._stub.DropAlias.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def alter_alias(self, collection_name, alias, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.alter_alias_request(collection_name, alias)
        rf = self._stub.AlterAlias.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def create_index(self, collection_name, field_name, params, timeout=None, **kwargs):
        # for historical reason, index_name contained in kwargs.
        index_name = kwargs.pop("index_name", DefaultConfigs.IndexName)
        copy_kwargs = copy.deepcopy(kwargs)

        collection_desc = self.describe_collection(collection_name, timeout=timeout, **copy_kwargs)

        valid_field = False
        for fields in collection_desc["fields"]:
            if field_name != fields["name"]:
                continue
            valid_field = True
            if fields["type"] != DataType.FLOAT_VECTOR and fields["type"] != DataType.BINARY_VECTOR:
                break
            # check index params on vector field.
            check_index_params(params)

        if not valid_field:
            raise MilvusException(message=f"cannot create index on non-existed field: {field_name}")

        # sync flush
        _async = kwargs.get("_async", False)
        kwargs["_async"] = False

        index_param = Prepare.create_index_request(collection_name, field_name, params, index_name=index_name)
        future = self._stub.CreateIndex.future(index_param, timeout=timeout)

        if _async:
            def _check():
                if kwargs.get("sync", True):
                    index_success, fail_reason = self.wait_for_creating_index(collection_name=collection_name,
                                                                              index_name=index_name,
                                                                              timeout=timeout, field_name=field_name)
                    if not index_success:
                        raise MilvusException(message=fail_reason)

            index_future = CreateIndexFuture(future)
            index_future.add_callback(_check)
            user_cb = kwargs.get("_callback", None)
            if user_cb:
                index_future.add_callback(user_cb)
            return index_future

        status = future.result()

        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

        if kwargs.get("sync", True):
            index_success, fail_reason = self.wait_for_creating_index(collection_name=collection_name,
                                                                      index_name=index_name,
                                                                      timeout=timeout, field_name=field_name)
            if not index_success:
                raise MilvusException(message=fail_reason)

        return Status(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def list_indexes(self, collection_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.describe_index_request(collection_name, "")

        rf = self._stub.DescribeIndex.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            return response.index_descriptions
        if status.error_code == Status.INDEX_NOT_EXIST:
            return []
        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def describe_index(self, collection_name, index_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        request = Prepare.describe_index_request(collection_name, index_name)

        rf = self._stub.DescribeIndex.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == Status.INDEX_NOT_EXIST:
            return None
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)
        if len(response.index_descriptions) == 1:
            info_dict = {kv.key: kv.value for kv in response.index_descriptions[0].params}
            info_dict['field_name'] = response.index_descriptions[0].field_name
            info_dict['index_name'] = response.index_descriptions[0].index_name
            if info_dict.get("params", None):
                info_dict["params"] = json.loads(info_dict["params"])
            return info_dict

        raise AmbiguousIndexName(message=ExceptionsMessage.AmbiguousIndexName)

    @retry_on_rpc_failure()
    def get_index_build_progress(self, collection_name, index_name, timeout=None):
        request = Prepare.describe_index_request(collection_name, index_name)
        rf = self._stub.DescribeIndex.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            if len(response.index_descriptions) == 1:
                index_desc = response.index_descriptions[0]
                return {'total_rows': index_desc.total_rows, 'indexed_rows': index_desc.indexed_rows}
            raise AmbiguousIndexName(message=ExceptionsMessage.AmbiguousIndexName)
        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def get_index_state(self, collection_name: str, index_name: str, timeout=None, **kwargs):
        request = Prepare.describe_index_request(collection_name, index_name)
        rf = self._stub.DescribeIndex.future(request, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

        if len(response.index_descriptions) == 1:
            index_desc = response.index_descriptions[0]
            return index_desc.state, index_desc.index_state_fail_reason
        # just for create_index.
        field_name = kwargs.pop("field_name", "")
        if field_name != "":
            for index_desc in response.index_descriptions:
                if index_desc.field_name == field_name:
                    return index_desc.state, index_desc.index_state_fail_reason

        raise AmbiguousIndexName(message=ExceptionsMessage.AmbiguousIndexName)

    @retry_on_rpc_failure()
    def wait_for_creating_index(self, collection_name, index_name, timeout=None, **kwargs):
        start = time.time()
        while True:
            time.sleep(0.5)
            state, fail_reason = self.get_index_state(collection_name, index_name, timeout=timeout, **kwargs)
            if state == IndexState.Finished:
                return True, fail_reason
            if state == IndexState.Failed:
                return False, fail_reason
            end = time.time()
            if isinstance(timeout, int) and end - start > timeout:
                raise MilvusException(message=f"collection {collection_name} create index {index_name} timeout in {timeout}s")

    @retry_on_rpc_failure()
    def load_collection(self, collection_name, replica_number=1, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name, replica_number=replica_number)
        _refresh = kwargs.get("_refresh", False)
        _resource_groups = kwargs.get("_resource_groups")
        request = Prepare.load_collection("", collection_name, replica_number, _refresh, _resource_groups)
        rf = self._stub.LoadCollection.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)
        _async = kwargs.get("_async", False)
        if not _async:
            self.wait_for_loading_collection(collection_name, timeout)

    @retry_on_rpc_failure()
    def load_collection_progress(self, collection_name, timeout=None):
        """ Return loading progress of collection """
        progress = self.get_loading_progress(collection_name, timeout=timeout)
        return {
            "loading_progress": f"{progress:.0f}%",
        }

    @retry_on_rpc_failure()
    def wait_for_loading_collection(self, collection_name, timeout=None):
        start = time.time()

        def can_loop(t) -> bool:
            return True if timeout is None else t <= (start + timeout)

        while can_loop(time.time()):
            progress = self.get_loading_progress(collection_name, timeout=timeout)
            if progress >= 100:
                return
            time.sleep(DefaultConfigs.WaitTimeDurationWhenLoad)
        raise MilvusException(message=f"wait for loading collection timeout, collection: {collection_name}")

    @retry_on_rpc_failure()
    def release_collection(self, collection_name, timeout=None):
        check_pass_param(collection_name=collection_name)
        request = Prepare.release_collection("", collection_name)
        rf = self._stub.ReleaseCollection.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def load_partitions(self, collection_name, partition_names, replica_number=1, timeout=None, **kwargs):
        check_pass_param(
            collection_name=collection_name,
            partition_name_array=partition_names,
            replica_number=replica_number)
        _refresh = kwargs.get("_refresh", False)
        _resource_groups = kwargs.get("_resource_groups")
        request = Prepare.load_partitions("", collection_name, partition_names, replica_number, _refresh,
                                          _resource_groups)
        future = self._stub.LoadPartitions.future(request, timeout=timeout)

        if kwargs.get("_async", False):
            def _check():
                if kwargs.get("sync", True):
                    self.wait_for_loading_partitions(collection_name, partition_names)

            load_partitions_future = LoadPartitionsFuture(future)
            load_partitions_future.add_callback(_check)

            user_cb = kwargs.get("_callback", None)
            if user_cb:
                load_partitions_future.add_callback(user_cb)

            return load_partitions_future

        response = future.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)
        sync = kwargs.get("sync", True)
        if sync:
            self.wait_for_loading_partitions(collection_name, partition_names)

    @retry_on_rpc_failure()
    def wait_for_loading_partitions(self, collection_name, partition_names, timeout=None):
        start = time.time()

        def can_loop(t) -> bool:
            return True if timeout is None else t <= (start + timeout)

        while can_loop(time.time()):
            progress = self.get_loading_progress(collection_name, partition_names, timeout=timeout)
            if progress >= 100:
                return
            time.sleep(DefaultConfigs.WaitTimeDurationWhenLoad)
        raise MilvusException(message=f"wait for loading partition timeout, collection: {collection_name}, partitions: {partition_names}")

    @retry_on_rpc_failure()
    def get_loading_progress(self, collection_name, partition_names=None, timeout=None):
        request = Prepare.get_loading_progress(collection_name, partition_names)
        response = self._stub.GetLoadingProgress.future(request, timeout=timeout).result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)
        return response.progress

    @retry_on_rpc_failure()
    def get_load_state(self, collection_name, partition_names=None, timeout=None):
        request = Prepare.get_load_state(collection_name, partition_names)
        response = self._stub.GetLoadState.future(request, timeout=timeout).result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)
        return LoadState(response.state)

    @retry_on_rpc_failure()
    def load_partitions_progress(self, collection_name, partition_names, timeout=None):
        """ Return loading progress of partitions """
        progress = self.get_loading_progress(collection_name, partition_names, timeout)
        return {
            "loading_progress": f"{progress:.0f}%",
        }

    @retry_on_rpc_failure()
    def release_partitions(self, collection_name, partition_names, timeout=None):
        check_pass_param(collection_name=collection_name, partition_name_array=partition_names)
        request = Prepare.release_partitions("", collection_name, partition_names)
        rf = self._stub.ReleasePartitions.future(request, timeout=timeout)
        response = rf.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def get_collection_stats(self, collection_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name)
        index_param = Prepare.get_collection_stats_request(collection_name)
        future = self._stub.GetCollectionStatistics.future(index_param, timeout=timeout)
        response = future.result()
        status = response.status
        if status.error_code == 0:
            return response.stats

        raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def get_flush_state(self, segment_ids, timeout=None, **kwargs):
        req = Prepare.get_flush_state_request(segment_ids)
        future = self._stub.GetFlushState.future(req, timeout=timeout)
        response = future.result()
        status = response.status
        if status.error_code == 0:
            return response.flushed  # todo: A wrapper class of PersistentSegmentInfo
        raise MilvusException(status.error_code, status.reason)

    # TODO seem not in use
    @retry_on_rpc_failure()
    def get_persistent_segment_infos(self, collection_name, timeout=None, **kwargs):
        req = Prepare.get_persistent_segment_info_request(collection_name)
        future = self._stub.GetPersistentSegmentInfo.future(req, timeout=timeout)
        response = future.result()
        status = response.status
        if status.error_code == 0:
            return response.infos  # todo: A wrapper class of PersistentSegmentInfo
        raise MilvusException(status.error_code, status.reason)

    def _wait_for_flushed(self, segment_ids, timeout=None, **kwargs):
        flush_ret = False
        start = time.time()
        while not flush_ret:
            flush_ret = self.get_flush_state(segment_ids, timeout, **kwargs)
            end = time.time()
            if timeout is not None:
                if end - start > timeout:
                    raise MilvusException(message=f"wait for flush timeout, segment ids: {segment_ids}")

            if not flush_ret:
                time.sleep(0.5)

    @retry_on_rpc_failure()
    def flush(self, collection_names: list, timeout=None, **kwargs):
        if collection_names in (None, []) or not isinstance(collection_names, list):
            raise ParamError(message="Collection name list can not be None or empty")

        for name in collection_names:
            check_pass_param(collection_name=name)

        request = Prepare.flush_param(collection_names)
        future = self._stub.Flush.future(request, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        def _check():
            for collection_name in collection_names:
                segment_ids = future.result().coll_segIDs[collection_name].data
                self._wait_for_flushed(segment_ids)

        if kwargs.get("_async", False):
            flush_future = FlushFuture(future)
            flush_future.add_callback(_check)

            user_cb = kwargs.get("_callback", None)
            if user_cb:
                flush_future.add_callback(user_cb)

            return flush_future

        _check()

    @retry_on_rpc_failure()
    def drop_index(self, collection_name, field_name, index_name, timeout=None, **kwargs):
        check_pass_param(collection_name=collection_name, field_name=field_name)
        request = Prepare.drop_index_request(collection_name, field_name, index_name)
        future = self._stub.DropIndex.future(request, timeout=timeout)
        response = future.result()
        if response.error_code != 0:
            raise MilvusException(response.error_code, response.reason)

    @retry_on_rpc_failure()
    def dummy(self, request_type, timeout=None, **kwargs):
        request = Prepare.dummy_request(request_type)
        future = self._stub.Dummy.future(request, timeout=timeout)
        return future.result()

    # TODO seems not in use
    @retry_on_rpc_failure()
    def fake_register_link(self, timeout=None):
        request = Prepare.register_link_request()
        future = self._stub.RegisterLink.future(request, timeout=timeout)
        return future.result().status

    # TODO seems not in use
    @retry_on_rpc_failure()
    def get(self, collection_name, ids, output_fields=None, partition_names=None, timeout=None):
        # TODO: some check
        request = Prepare.retrieve_request(collection_name, ids, output_fields, partition_names)
        future = self._stub.Retrieve.future(request, timeout=timeout)
        return future.result()

    @retry_on_rpc_failure()
    def query(self, collection_name, expr, output_fields=None, partition_names=None, timeout=None, **kwargs):
        if output_fields is not None and not isinstance(output_fields, (list,)):
            raise ParamError(message="Invalid query format. 'output_fields' must be a list")
        collection_schema = kwargs.get("schema", None)
        if not collection_schema:
            collection_schema = self.describe_collection(collection_name, timeout)
        consistency_level = collection_schema["consistency_level"]
        # overwrite the consistency level defined when user created the collection
        consistency_level = get_consistency_level(kwargs.get("consistency_level", consistency_level))

        ts_utils.construct_guarantee_ts(consistency_level, collection_name, kwargs)
        request = Prepare.query_request(collection_name, expr, output_fields, partition_names, **kwargs)

        future = self._stub.Query.future(request, timeout=timeout)
        response = future.result()
        if response.status.error_code == Status.EMPTY_COLLECTION:
            return []
        if response.status.error_code != Status.SUCCESS:
            raise MilvusException(response.status.error_code, response.status.reason)

        num_fields = len(response.fields_data)
        # check has fields
        if num_fields == 0:
            raise MilvusException(message="No fields returned")

        # check if all lists are of the same length
        it = iter(response.fields_data)
        num_entities = len_of(next(it))
        if not all(len_of(field_data) == num_entities for field_data in it):
            raise MilvusException(message="The length of fields data is inconsistent")

        # transpose
        results = []
        for index in range(0, num_entities):
            result = {}
            for field_data in response.fields_data:
                if field_data.type == DataType.BOOL:
                    result[field_data.field_name] = field_data.scalars.bool_data.data[index]
                elif field_data.type in (DataType.INT8, DataType.INT16, DataType.INT32):
                    result[field_data.field_name] = field_data.scalars.int_data.data[index]
                elif field_data.type == DataType.INT64:
                    result[field_data.field_name] = field_data.scalars.long_data.data[index]
                elif field_data.type == DataType.FLOAT:
                    result[field_data.field_name] = np.single(field_data.scalars.float_data.data[index])
                elif field_data.type == DataType.DOUBLE:
                    result[field_data.field_name] = field_data.scalars.double_data.data[index]
                elif field_data.type == DataType.VARCHAR:
                    result[field_data.field_name] = field_data.scalars.string_data.data[index]
                elif field_data.type == DataType.STRING:
                    raise MilvusException(message="Not support string yet")
                    # result[field_data.field_name] = field_data.scalars.string_data.data[index]
                elif field_data.type == DataType.FLOAT_VECTOR:
                    dim = field_data.vectors.dim
                    start_pos = index * dim
                    end_pos = index * dim + dim
                    result[field_data.field_name] = [np.single(x) for x in
                                                     field_data.vectors.float_vector.data[start_pos:end_pos]]
                elif field_data.type == DataType.BINARY_VECTOR:
                    dim = field_data.vectors.dim
                    start_pos = index * (int(dim / 8))
                    end_pos = (index + 1) * (int(dim / 8))
                    result[field_data.field_name] = field_data.vectors.binary_vector[start_pos:end_pos]
            results.append(result)

        return results

    @retry_on_rpc_failure()
    def load_balance(self, collection_name: str, src_node_id, dst_node_ids, sealed_segment_ids, timeout=None, **kwargs):
        req = Prepare.load_balance_request(collection_name, src_node_id, dst_node_ids, sealed_segment_ids)
        future = self._stub.LoadBalance.future(req, timeout=timeout)
        status = future.result()
        if status.error_code != 0:
            raise MilvusException(status.error_code, status.reason)

    @retry_on_rpc_failure()
    def compact(self, collection_name, timeout=None, **kwargs) -> int:
        request = Prepare.describe_collection_request(collection_name)
        rf = self._stub.DescribeCollection.future(request, timeout=timeout)
        response = rf.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        req = Prepare.manual_compaction(response.collectionID, 0)
        future = self._stub.ManualCompaction.future(req, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        return response.compactionID

    @retry_on_rpc_failure()
    def get_compaction_state(self, compaction_id, timeout=None, **kwargs) -> CompactionState:
        req = Prepare.get_compaction_state(compaction_id)

        future = self._stub.GetCompactionState.future(req, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        return CompactionState(
            compaction_id,
            State.new(response.state),
            response.executingPlanNo,
            response.timeoutPlanNo,
            response.completedPlanNo
        )

    @retry_on_rpc_failure()
    def wait_for_compaction_completed(self, compaction_id, timeout=None, **kwargs):
        start = time.time()
        while True:
            time.sleep(0.5)
            compaction_state = self.get_compaction_state(compaction_id, timeout, **kwargs)
            if compaction_state.state == State.Completed:
                return True
            if compaction_state == State.UndefiedState:
                return False
            end = time.time()
            if timeout is not None:
                if end - start > timeout:
                    raise MilvusException(message=f"get compaction state timeout, compaction id: {compaction_id}")

    @retry_on_rpc_failure()
    def get_compaction_plans(self, compaction_id, timeout=None, **kwargs) -> CompactionPlans:
        req = Prepare.get_compaction_state_with_plans(compaction_id)

        future = self._stub.GetCompactionStateWithPlans.future(req, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        cp = CompactionPlans(compaction_id, response.state)

        cp.plans = [Plan(m.sources, m.target) for m in response.mergeInfos]

        return cp

    @retry_on_rpc_failure()
    def get_replicas(self, collection_name, timeout=None, **kwargs) -> Replica:
        collection_id = self.describe_collection(collection_name, timeout, **kwargs)["collection_id"]

        req = Prepare.get_replicas(collection_id)
        future = self._stub.GetReplicas.future(req, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        groups = []
        for replica in response.replicas:
            shards = [Shard(s.dm_channel_name, s.node_ids, s.leaderID) for s in replica.shard_replicas]
            groups.append(Group(replica.replicaID, shards, replica.node_ids, replica.resource_group_name,
                                replica.num_outbound_node))

        return Replica(groups)

    @retry_on_rpc_failure()
    def do_bulk_insert(self, collection_name, partition_name, files: list, timeout=None, **kwargs) -> int:
        req = Prepare.do_bulk_insert(collection_name, partition_name, files, **kwargs)
        future = self._stub.Import.future(req, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)
        if len(response.tasks) == 0:
            raise MilvusException(common_pb2.UNEXPECTED_ERROR, "no task id returned from server")
        return response.tasks[0]

    @retry_on_rpc_failure()
    def get_bulk_insert_state(self, task_id, timeout=None, **kwargs) -> BulkInsertState:
        req = Prepare.get_bulk_insert_state(task_id)
        future = self._stub.GetImportState.future(req, timeout=timeout)
        resp = future.result()
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        state = BulkInsertState(task_id, resp.state, resp.row_count, resp.id_list, resp.infos, resp.create_ts)
        return state

    @retry_on_rpc_failure()
    def list_bulk_insert_tasks(self, limit, collection_name, timeout=None, **kwargs) -> list:
        req = Prepare.list_bulk_insert_tasks(limit, collection_name)
        future = self._stub.ListImportTasks.future(req, timeout=timeout)
        resp = future.result()
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)

        tasks = [BulkInsertState(t.id, t.state, t.row_count, t.id_list, t.infos, t.create_ts)
                 for t in resp.tasks]
        return tasks

    @retry_on_rpc_failure()
    def create_user(self, user, password, timeout=None, **kwargs):
        check_pass_param(user=user, password=password)
        req = Prepare.create_user_request(user, password)
        resp = self._stub.CreateCredential(req, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def update_password(self, user, old_password, new_password, timeout=None, **kwargs):
        req = Prepare.update_password_request(user, old_password, new_password)
        resp = self._stub.UpdateCredential(req, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def delete_user(self, user, timeout=None, **kwargs):
        req = Prepare.delete_user_request(user)
        resp = self._stub.DeleteCredential(req, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def list_usernames(self, timeout=None, **kwargs):
        req = Prepare.list_usernames_request()
        resp = self._stub.ListCredUsers(req, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return resp.usernames

    @retry_on_rpc_failure()
    def create_role(self, role_name, timeout=None, **kwargs):
        req = Prepare.create_role_request(role_name)
        resp = self._stub.CreateRole(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def drop_role(self, role_name, timeout=None, **kwargs):
        req = Prepare.drop_role_request(role_name)
        resp = self._stub.DropRole(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def add_user_to_role(self, username, role_name, timeout=None, **kwargs):
        req = Prepare.operate_user_role_request(username, role_name, milvus_types.OperateUserRoleType.AddUserToRole)
        resp = self._stub.OperateUserRole(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def remove_user_from_role(self, username, role_name, timeout=None, **kwargs):
        req = Prepare.operate_user_role_request(username, role_name,
                                                milvus_types.OperateUserRoleType.RemoveUserFromRole)
        resp = self._stub.OperateUserRole(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def select_one_role(self, role_name, include_user_info, timeout=None, **kwargs):
        req = Prepare.select_role_request(role_name, include_user_info)
        resp = self._stub.SelectRole(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return RoleInfo(resp.results)

    @retry_on_rpc_failure()
    def select_all_role(self, include_user_info, timeout=None, **kwargs):
        req = Prepare.select_role_request(None, include_user_info)
        resp = self._stub.SelectRole(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return RoleInfo(resp.results)

    @retry_on_rpc_failure()
    def select_one_user(self, username, include_role_info, timeout=None, **kwargs):
        req = Prepare.select_user_request(username, include_role_info)
        resp = self._stub.SelectUser(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return UserInfo(resp.results)

    @retry_on_rpc_failure()
    def select_all_user(self, include_role_info, timeout=None, **kwargs):
        req = Prepare.select_user_request(None, include_role_info)
        resp = self._stub.SelectUser(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return UserInfo(resp.results)

    @retry_on_rpc_failure()
    def grant_privilege(self, role_name, object, object_name, privilege, timeout=None, **kwargs):
        req = Prepare.operate_privilege_request(role_name, object, object_name, privilege,
                                                milvus_types.OperatePrivilegeType.Grant)
        resp = self._stub.OperatePrivilege(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def revoke_privilege(self, role_name, object, object_name, privilege, timeout=None, **kwargs):
        req = Prepare.operate_privilege_request(role_name, object, object_name, privilege,
                                                milvus_types.OperatePrivilegeType.Revoke)
        resp = self._stub.OperatePrivilege(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def select_grant_for_one_role(self, role_name, timeout=None, **kwargs):
        req = Prepare.select_grant_request(role_name, None, None)
        resp = self._stub.SelectGrant(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)

        return GrantInfo(resp.entities)

    @retry_on_rpc_failure()
    def select_grant_for_role_and_object(self, role_name, object, object_name, timeout=None, **kwargs):
        req = Prepare.select_grant_request(role_name, object, object_name)
        resp = self._stub.SelectGrant(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)

        return GrantInfo(resp.entities)

    @retry_on_rpc_failure()
    def get_server_version(self, timeout=None, **kwargs) -> str:
        req = Prepare.get_server_version()
        resp = self._stub.GetVersion(req, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)

        return resp.version

    @retry_on_rpc_failure()
    def create_resource_group(self, name, timeout=None, **kwargs):
        req = Prepare.create_resource_group(name)
        resp = self._stub.CreateResourceGroup(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def drop_resource_group(self, name, timeout=None, **kwargs):
        req = Prepare.drop_resource_group(name)
        resp = self._stub.DropResourceGroup(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def list_resource_groups(self, timeout=None, **kwargs):
        req = Prepare.list_resource_groups()
        resp = self._stub.ListResourceGroups(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return list(resp.resource_groups)

    @retry_on_rpc_failure()
    def describe_resource_group(self, name, timeout=None, **kwargs) -> ResourceGroupInfo:
        req = Prepare.describe_resource_group(name)
        resp = self._stub.DescribeResourceGroup(req, wait_for_ready=True, timeout=timeout)
        if resp.status.error_code != 0:
            raise MilvusException(resp.status.error_code, resp.status.reason)
        return ResourceGroupInfo(resp.resource_group)

    @retry_on_rpc_failure()
    def transfer_node(self, source, target, num_node, timeout=None, **kwargs):
        req = Prepare.transfer_node(source, target, num_node)
        resp = self._stub.TransferNode(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def transfer_replica(self, source, target, collection_name, num_replica, timeout=None, **kwargs):
        req = Prepare.transfer_replica(source, target, collection_name, num_replica)
        resp = self._stub.TransferReplica(req, wait_for_ready=True, timeout=timeout)
        if resp.error_code != 0:
            raise MilvusException(resp.error_code, resp.reason)

    @retry_on_rpc_failure()
    def get_flush_all_state(self, flush_all_ts, timeout=None, **kwargs):
        req = Prepare.get_flush_all_state_request(flush_all_ts)
        response = self._stub.GetFlushAllState(req, timeout=timeout)
        status = response.status
        if status.error_code == 0:
            return response.flushed
        raise MilvusException(status.error_code, status.reason)

    def _wait_for_flush_all(self, flush_all_ts, timeout=None, **kwargs):
        flush_ret = False
        start = time.time()
        while not flush_ret:
            flush_ret = self.get_flush_all_state(flush_all_ts, timeout, **kwargs)
            end = time.time()
            if timeout is not None:
                if end - start > timeout:
                    raise MilvusException(message=f"wait for flush all timeout, flush_all_ts: {flush_all_ts}")

            if not flush_ret:
                time.sleep(5)

    @retry_on_rpc_failure()
    def flush_all(self, timeout=None, **kwargs):
        request = Prepare.flush_all_request()
        future = self._stub.FlushAll.future(request, timeout=timeout)
        response = future.result()
        if response.status.error_code != 0:
            raise MilvusException(response.status.error_code, response.status.reason)

        def _check():
            self._wait_for_flush_all(response.flush_all_ts, timeout, **kwargs)

        if kwargs.get("_async", False):
            flush_future = FlushFuture(future)
            flush_future.add_callback(_check)

            user_cb = kwargs.get("_callback", None)
            if user_cb:
                flush_future.add_callback(user_cb)

            return flush_future

        _check()



from pymilvus import utility, connections, Collection, CollectionSchema, FieldSchema, DataType
from pymilvus.orm.role import Role

import random
from sklearn import preprocessing

_CONNECTION = "demo"
_FOO_CONNECTION = "foo_connection"
_HOST = '127.0.0.1'
_PORT = '19530'
_ROOT = "root"
_ROOT_PASSWORD = "Milvus"
_COLLECTION_NAME = "foocol2"


def connect_to_milvus(connection=_CONNECTION, user=_ROOT, password=_ROOT_PASSWORD):
    print(f"connect to milvus\n")
    connections.connect(alias=connection,
                        host=_HOST,
                        port=_PORT,
                        user=user,
                        password=password,
                        )


def create_credential(user, password, connection=_CONNECTION):
    print(f"create credential, user: {user}, password: {password}")
    utility.create_user(user, password, using=connection)
    print(f"create credential down\n")


def update_credential(user, password, old_password, connection=_CONNECTION):
    print(f"update credential, user: {user}, password: {password}, old password: {old_password}")
    utility.reset_password(user, old_password, password, using=connection)
    print(f"update credential down\n")


def drop_credential(user, connection=_CONNECTION):
    print(f"drop credential, user: {user}")
    utility.delete_user(user, using=connection)
    print(f"drop credential down\n")


def select_one_user(username, connection=_CONNECTION):
    print(f"select one user, username: {username}")
    roles = utility.list_user(username, True, using=connection)
    print(roles)
    print(f"select one user done\n")


def select_all_user(connection=_CONNECTION):
    print(f"select all user")
    userinfo = utility.list_users(False, using=connection)
    print(userinfo)
    userinfo = utility.list_users(True, using=connection)
    print(userinfo)
    print(f"select all user done\n")


def select_all_role(connection=_CONNECTION):
    print(f"select_all_role")
    roles = utility.list_roles(False, using=connection)
    print(roles)
    roles = utility.list_roles(True, using=connection)
    print(roles)
    print(f"select_all_role done\n")


def has_collection(collection_name, connection=_CONNECTION):
    print(f"has collection, collection_name: {collection_name}")
    has = utility.has_collection("hello_milvus", using=connection)
    print(has)
    print(f"has collection end")


default_dim = 128
default_nb = 1000


def gen_float_vectors(num, dim, is_normal=True):
    vectors = [[random.random() for _ in range(dim)] for _ in range(num)]
    vectors = preprocessing.normalize(vectors, axis=1, norm='l2')
    return vectors.tolist()


def gen_float_data(nb, is_normal=False):
    vectors = gen_float_vectors(nb, default_dim, is_normal)
    entities = [
        [i for i in range(nb)],
        [float(i) for i in range(nb)],
        vectors
    ]
    return entities


# rbac I
def rbac_collection(connection=_CONNECTION):
    print(f"rbac_collection")
    default_float_vec_field_name = "float_vector"
    default_fields = [
        FieldSchema(name="int64", dtype=DataType.INT64, is_primary=True),
        FieldSchema(name="double", dtype=DataType.DOUBLE),
        FieldSchema(name=default_float_vec_field_name, dtype=DataType.FLOAT_VECTOR, dim=default_dim)
    ]
    default_schema = CollectionSchema(fields=default_fields, description="create_collection_rbac")
    collection = Collection(name=_COLLECTION_NAME, schema=default_schema, using=connection)

    data = gen_float_data(default_nb)
    collection.insert(data)

    is_exception = False
    try:
        collection.drop()
    except Exception as e:
        print(e)
        is_exception = True
    assert is_exception

    collection = Collection(name=_COLLECTION_NAME, schema=default_schema, using=_CONNECTION)
    collection.drop()
    print(f"rbac_collection done")


# rbac II
def rbac_user(username, password, role_name, connection=_CONNECTION):
    update_credential(username, "pfoo1235", password, connection=connection)
    print(select_one_user(username, connection=connection))
    is_exception = False
    try:
        select_all_user(connection=connection)
    except Exception as e:
        print(e)
        is_exception = True
    assert is_exception
    role = Role(role_name, using=_CONNECTION)
    role.grant("User", "*", "SelectUser")
    print(select_all_user(connection))
    role.revoke("User", "*", "SelectUser")


def role_example():
    role_name = "role_test5"
    role = Role(role_name, using=_CONNECTION)
    print(f"create role, role_name: {role_name}")
    role.create()
    print(f"get users")
    role.get_users()
    print(f"select all role")
    print(select_all_role())
    print(f"drop role")
    role.drop()


def associate_users_with_roles_example():
    username = "root"
    role_name = "public"
    role = Role(role_name, using=_CONNECTION)
    print(f"add user")
    role.add_user(username)
    print(f"get users")
    role.get_users()
    print(select_one_user(username))
    print(select_all_user())
    print(f"remove user")
    role.remove_user(username)


def privilege_example():
    print(f"privilege example")

    username = "foo53"
    password = "pfoo123"
    role_name = "general53"
    privilege_create = "CreateCollection"
    privilege_insert = "Insert"
    object_name = _COLLECTION_NAME

    create_credential(username, password)
    role = Role(role_name, using=_CONNECTION)
    print(f"create role, role_name: {role_name}")
    role.create()
    print(f"add user")
    role.add_user(username)
    print(f"grant privilege")
    role.grant("Global", "*", privilege_create)
    role.grant("Collection", object_name, privilege_insert)
    # role.grant("Collection", object_name, "*")
    # role.grant("Collection", "*", privilege_insert)

    print(f"list grants")
    print(role.list_grants())
    print(f"list grant")
    print(role.list_grant("Collection", object_name))

    connect_to_milvus(connection=_FOO_CONNECTION, user=username, password=password)
    has_collection(_COLLECTION_NAME, connection=_FOO_CONNECTION)
    rbac_collection(connection=_FOO_CONNECTION)
    rbac_user(username, password, role_name, connection=_FOO_CONNECTION)

    print(f"revoke privilege")
    role.revoke("Global", "*", privilege_create)
    role.revoke("Collection", object_name, privilege_insert)
    # role.revoke("Collection", object_name, "*")
    # role.revoke("Collection", "*", privilege_insert)
    print(f"remove user")
    role.remove_user(username)
    role.drop()
    drop_credential(username)


def run():
    connect_to_milvus()
    role_example()
    associate_users_with_roles_example()
    privilege_example()


if __name__ == "__main__":
    run()



import random

from pymilvus import Milvus, DataType

if __name__ == "__main__":
    c = Milvus("localhost", "19530")

    collection_name = f"test_{random.randint(10000, 99999)}"

    c.create_collection(collection_name, {"fields": [
        {
            "name": "f1",
            "type": DataType.FLOAT_VECTOR,
            "metric_type": "L2",
            "params": {"dim": 4},
            "indexes": [{"metric_type": "L2"}]
        },
        {
            "name": "age",
            "type": DataType.FLOAT,
        },
        {
            "name": "id",
            "type": DataType.INT64,
            "auto_id": True,
            "is_primary": True,
        }
    ],
    }, orm=True)

    assert c.has_collection(collection_name)

    ids = c.insert(collection_name, [
        {"name": "f1", "type": DataType.FLOAT_VECTOR, "values": [[1.1, 2.2, 3.3, 4.4], [5.5, 6.6, 7.7, 8.8]]},
        {"name": "age", "type": DataType.FLOAT, "values": [3.45, 8.9]}
    ], orm=True)

    c.flush([collection_name])

    c.load_collection(collection_name)

    #############################################################
    search_params = {"metric_type": "L2", "params": {"nprobe": 1}}

    results = c.search(collection_name, [[1.1, 2.2, 3.3, 4.4]],
                       "f1", param=search_params, limit=2, output_fields=["id"])

    print("search results: ", results[0][0].entity, " + ", results[0][1].entity)
    #
    # print("Test entity.get: ", results[0][0].entity.get("age"))
    # print("Test entity.value_of_field: ", results[0][0].entity.value_of_field("age"))
    # print("Test entity.fields: ", results[0][0].entity.fields)
    #############################################################

    ids_expr = ",".join(str(x) for x in ids.primary_keys)

    expr = "id in [ " + ids_expr + " ] "

    print(expr)

    res = c.query(collection_name, expr)
    print(res)



import random

from pymilvus import (
    connections,
    FieldSchema, CollectionSchema, DataType,
    Collection,
    utility
)

# This example shows how to:
#   1. connect to Milvus server
#   2. create a collection
#   3. insert entities
#   4. create index
#   5. search


_HOST = '127.0.0.1'
_PORT = '19530'

# Const names
_COLLECTION_NAME = 'demo'
_ID_FIELD_NAME = 'id_field'
_VECTOR_FIELD_NAME = 'float_vector_field'

# Vector parameters
_DIM = 128
_INDEX_FILE_SIZE = 32  # max file size of stored index

# Scalar
_ATTR1_NAME = "attr1"
_ATTR2_NAME = "attr2"

# Index parameters
_METRIC_TYPE = 'L2'
_INDEX_TYPE = 'IVF_FLAT'
_NLIST = 1024
_NPROBE = 16
_TOPK = 3


# Create a Milvus connection
def create_connection():
    print(f"\nCreate connection...")
    connections.connect(host=_HOST, port=_PORT)
    print(f"\nList connections:")
    print(connections.list_connections())


# Create a collection named 'demo'
def create_collection(name, id_field, vector_field, attr1_name, attr2_name):
    field1 = FieldSchema(name=id_field, dtype=DataType.VARCHAR, description="varchar", is_primary=True, max_length=10)
    field2 = FieldSchema(name=vector_field, dtype=DataType.FLOAT_VECTOR, description="float vector", dim=_DIM)
    field3 = FieldSchema(name=attr1_name, dtype=DataType.INT64, description="attr1")
    field4 = FieldSchema(name=attr2_name, dtype=DataType.DOUBLE, description="attr2")
    schema = CollectionSchema(fields=[field1, field2, field3, field4])
    collection = Collection(name=name, data=None, schema=schema)
    print("\ncollection created:", name)
    return collection


def has_collection(name):
    return utility.has_collection(name)


# Drop a collection in Milvus
def drop_collection(name):
    utility.drop_collection(name)
    print("\nDrop collection: {}".format(name))


# List all collections in Milvus
def list_collections():
    print("\nlist collections:")
    print(utility.list_collections())


def insert(collection, num, dim):
    data = [
        [str(i) for i in range(num)],
        [[random.random() for _ in range(dim)] for _ in range(num)],
        [random.randint(1, 10000) for _ in range(num)],
        [random.random() for _ in range(num)],
    ]
    collection.insert(data)
    return data[1]


def get_entity_num(collection):
    print("\nThe number of entity:")
    print(collection.num_entities)


def create_index(collection, filed_name, index_param, index_name):
    collection.create_index(filed_name, index_param, index_name=index_name)
    index = collection.index(index_name=index_name)
    print(f"\nCreated index:\n{index.params}")


def drop_index(collection, index_name):
    collection.drop_index(index_name=index_name)
    print("\nDrop index sucessfully")


def load_collection(collection):
    collection.load()


def release_collection(collection):
    collection.release()


def search(collection, vector_field, id_field, search_vectors):
    search_param = {
        "data": search_vectors,
        "anns_field": vector_field,
        "param": {"metric_type": _METRIC_TYPE, "params": {"nprobe": _NPROBE}},
        "limit": _TOPK,
        "expr": 'id_field > "0"'}

    results = collection.search(**search_param)

    for i, result in enumerate(results):
        print(f"\nSearch result for {i}th vector: ")
        for j, res in enumerate(result):
            print(f"Top {j}: {res}")


def main():
    # create a connection
    create_connection()

    # drop collection if the collection exists
    if has_collection(_COLLECTION_NAME):
        drop_collection(_COLLECTION_NAME)

    # create collection
    collection = create_collection(_COLLECTION_NAME, _ID_FIELD_NAME, _VECTOR_FIELD_NAME, _ATTR1_NAME, _ATTR2_NAME)

    # show collections
    list_collections()

    # insert 10000 vectors with 128 dimension
    vectors = insert(collection, 10000, _DIM)

    collection.flush()
    # get the number of entities
    get_entity_num(collection)

    # create index
    vec_index_param = {
        "index_type": _INDEX_TYPE,
        "params": {"nlist": _NLIST},
        "metric_type": _METRIC_TYPE}
    scalar_index_param = {"index_type": "inverted_index"}  # TODO: replace this with real impl.

    vector_index_name = "vector_index"
    create_index(collection, _VECTOR_FIELD_NAME, vec_index_param, vector_index_name)
    print(f"has_index {vector_index_name}: ", collection.has_index())
    print("index: ", collection.index().to_dict())
    utility.wait_for_index_building_complete(collection.name)
    print("index building progress: ", utility.index_building_progress(collection.name))

    varchar_index_name = "varchar_id_index"
    create_index(collection, _ID_FIELD_NAME, scalar_index_param, varchar_index_name)
    print(f"has_index {varchar_index_name}: ", collection.has_index(index_name=varchar_index_name))
    print("all indexes:")
    for index in collection.indexes:
        print(index.to_dict())
    utility.wait_for_index_building_complete(collection.name, varchar_index_name)
    print("index building progress: ", utility.index_building_progress(collection.name, varchar_index_name))

    # load data to memory
    load_collection(collection)

    # search
    search(collection, _VECTOR_FIELD_NAME, _ID_FIELD_NAME, vectors[:3])

    # release memory
    release_collection(collection)

    # drop collection index
    drop_index(collection, vector_index_name)
    drop_index(collection, varchar_index_name)

    # drop collection
    drop_collection(_COLLECTION_NAME)


if __name__ == '__main__':
    main()



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

Abnormal: StateCode
AllocateSegment: MsgType
AlterAlias: MsgType
AlterCollection: MsgType
BinaryVector: PlaceholderType
BoolExprV1: DslType
Bounded: ConsistencyLevel
BuildIndexError: ErrorCode
CacheFailed: ErrorCode
CannotCreateFile: ErrorCode
CannotCreateFolder: ErrorCode
CannotDeleteFile: ErrorCode
CannotDeleteFolder: ErrorCode
Collection: ObjectType
CollectionNameNotFound: ErrorCode
CollectionNotExists: ErrorCode
Completed: CompactionState
ConnectFailed: ErrorCode
CreateAlias: MsgType
CreateCollection: MsgType
CreateCredential: MsgType
CreateCredentialFailure: ErrorCode
CreateIndex: MsgType
CreatePartition: MsgType
CreateResourceGroup: MsgType
CreateRole: MsgType
CreateRoleFailure: ErrorCode
Customized: ConsistencyLevel
DDRequestRace: ErrorCode
DESCRIPTOR: _descriptor.FileDescriptor
DataCoordNA: ErrorCode
DataNodeTt: MsgType
Delete: MsgType
DeleteCredential: MsgType
DeleteCredentialFailure: ErrorCode
DescribeCollection: MsgType
DescribeIndex: MsgType
DescribePartition: MsgType
DescribeResourceGroup: MsgType
DescribeSegment: MsgType
DescribeSegments: MsgType
DiskQuotaExhausted: ErrorCode
DropAlias: MsgType
DropCollection: MsgType
DropIndex: MsgType
DropPartition: MsgType
DropResourceGroup: MsgType
DropRole: MsgType
DropRoleFailure: ErrorCode
Dropped: SegmentState
Dsl: DslType
EmptyCollection: ErrorCode
Eventually: ConsistencyLevel
Executing: CompactionState
Failed: IndexState
FileNotFound: ErrorCode
Finished: IndexState
FloatVector: PlaceholderType
Flush: MsgType
Flushed: SegmentState
Flushing: SegmentState
ForceDeny: ErrorCode
GetCollectionStatistics: MsgType
GetCredential: MsgType
GetCredentialFailure: ErrorCode
GetDistribution: MsgType
GetIndexBuildProgress: MsgType
GetIndexState: MsgType
GetPartitionStatistics: MsgType
GetRecoveryInfo: MsgType
GetReplicas: MsgType
GetSegmentState: MsgType
GetShardLeaders: MsgType
GetSystemConfigs: MsgType
GetUserFailure: ErrorCode
Global: ObjectType
Growing: SegmentState
HandoffSegments: MsgType
HasCollection: MsgType
HasPartition: MsgType
Healthy: StateCode
IllegalArgument: ErrorCode
IllegalCollectionName: ErrorCode
IllegalDimension: ErrorCode
IllegalIndexType: ErrorCode
IllegalMetricType: ErrorCode
IllegalNLIST: ErrorCode
IllegalRowRecord: ErrorCode
IllegalSearchResult: ErrorCode
IllegalTOPK: ErrorCode
IllegalVectorID: ErrorCode
ImportCompleted: ImportState
ImportFailed: ImportState
ImportFailedAndCleaned: ImportState
ImportPending: ImportState
ImportPersisted: ImportState
ImportStarted: ImportState
Importing: SegmentState
InProgress: IndexState
IndexNotExist: ErrorCode
IndexStateNone: IndexState
Initializing: StateCode
Insert: MsgType
InsufficientMemoryToLoad: ErrorCode
ListCredUsernames: MsgType
ListCredUsersFailure: ErrorCode
ListPolicy: MsgType
ListPolicyFailure: ErrorCode
ListResourceGroups: MsgType
LoadBalanceSegments: MsgType
LoadCollection: MsgType
LoadIndex: MsgType
LoadPartitions: MsgType
LoadSegments: MsgType
LoadStateLoaded: LoadState
LoadStateLoading: LoadState
LoadStateNotExist: LoadState
LoadStateNotLoad: LoadState
MemoryQuotaExhausted: ErrorCode
MetaFailed: ErrorCode
NoReplicaAvailable: ErrorCode
NodeIDNotMatch: ErrorCode
None: PlaceholderType
NotExist: SegmentState
NotFoundTSafer: ErrorCode
NotReadyCoordActivating: ErrorCode
NotReadyServe: ErrorCode
NotShardLeader: ErrorCode
OperatePrivilege: MsgType
OperatePrivilegeFailure: ErrorCode
OperateUserRole: MsgType
OperateUserRoleFailure: ErrorCode
OutOfMemory: ErrorCode
PRIVILEGE_EXT_OBJ_FIELD_NUMBER: _ClassVar[int]
PermissionDenied: ErrorCode
PrivilegeAll: ObjectPrivilege
PrivilegeCompaction: ObjectPrivilege
PrivilegeCreateCollection: ObjectPrivilege
PrivilegeCreateIndex: ObjectPrivilege
PrivilegeCreateOwnership: ObjectPrivilege
PrivilegeCreateResourceGroup: ObjectPrivilege
PrivilegeDelete: ObjectPrivilege
PrivilegeDescribeCollection: ObjectPrivilege
PrivilegeDescribeResourceGroup: ObjectPrivilege
PrivilegeDropCollection: ObjectPrivilege
PrivilegeDropIndex: ObjectPrivilege
PrivilegeDropOwnership: ObjectPrivilege
PrivilegeDropResourceGroup: ObjectPrivilege
PrivilegeFlush: ObjectPrivilege
PrivilegeGetLoadState: ObjectPrivilege
PrivilegeGetLoadingProgress: ObjectPrivilege
PrivilegeGetStatistics: ObjectPrivilege
PrivilegeImport: ObjectPrivilege
PrivilegeIndexDetail: ObjectPrivilege
PrivilegeInsert: ObjectPrivilege
PrivilegeListResourceGroups: ObjectPrivilege
PrivilegeLoad: ObjectPrivilege
PrivilegeLoadBalance: ObjectPrivilege
PrivilegeManageOwnership: ObjectPrivilege
PrivilegeQuery: ObjectPrivilege
PrivilegeRelease: ObjectPrivilege
PrivilegeSearch: ObjectPrivilege
PrivilegeSelectOwnership: ObjectPrivilege
PrivilegeSelectUser: ObjectPrivilege
PrivilegeShowCollections: ObjectPrivilege
PrivilegeTransferNode: ObjectPrivilege
PrivilegeTransferReplica: ObjectPrivilege
PrivilegeUpdateUser: ObjectPrivilege
QueryNodeStats: MsgType
RateLimit: ErrorCode
RefreshPolicyInfoCache: MsgType
RefreshPolicyInfoCacheFailure: ErrorCode
ReleaseCollection: MsgType
ReleasePartitions: MsgType
ReleaseSegments: MsgType
RemoveDmChannels: MsgType
RemoveQueryChannels: MsgType
RenameCollection: MsgType
RequestID: MsgType
RequestTSO: MsgType
ResendSegmentStats: MsgType
Retrieve: MsgType
RetrieveResult: MsgType
Retry: IndexState
Sealed: SegmentState
SealedSegmentsChangeInfo: MsgType
Search: MsgType
SearchResult: MsgType
SegmentFlushDone: MsgType
SegmentInfo: MsgType
SegmentNotFound: ErrorCode
SegmentStateNone: SegmentState
SegmentStatistics: MsgType
SelectGrant: MsgType
SelectGrantFailure: ErrorCode
SelectResource: MsgType
SelectResourceFailure: ErrorCode
SelectRole: MsgType
SelectRoleFailure: ErrorCode
SelectUser: MsgType
SelectUserFailure: ErrorCode
Session: ConsistencyLevel
ShowCollections: MsgType
ShowPartitions: MsgType
ShowSegments: MsgType
StandBy: StateCode
Stopping: StateCode
Strong: ConsistencyLevel
Success: ErrorCode
SyncDistribution: MsgType
SystemInfo: MsgType
TimeTick: MsgType
TimeTickLongDelay: ErrorCode
TransferNode: MsgType
TransferReplica: MsgType
UndefiedState: CompactionState
Undefined: MsgType
UnexpectedError: ErrorCode
Unissued: IndexState
UnsubDmChannel: MsgType
UpdateCredential: MsgType
UpdateCredentialFailure: ErrorCode
UpdateImportTaskFailure: ErrorCode
UpsertAutoIDTrue: ErrorCode
User: ObjectType
WatchDeltaChannels: MsgType
WatchDmChannels: MsgType
WatchQueryChannels: MsgType
privilege_ext_obj: _descriptor.FieldDescriptor

class Address(_message.Message):
    __slots__ = ["ip", "port"]
    IP_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    ip: str
    port: int
    def __init__(self, ip: _Optional[str] = ..., port: _Optional[int] = ...) -> None: ...

class Blob(_message.Message):
    __slots__ = ["value"]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: bytes
    def __init__(self, value: _Optional[bytes] = ...) -> None: ...

class DMLMsgHeader(_message.Message):
    __slots__ = ["base", "shardName"]
    BASE_FIELD_NUMBER: _ClassVar[int]
    SHARDNAME_FIELD_NUMBER: _ClassVar[int]
    base: MsgBase
    shardName: str
    def __init__(self, base: _Optional[_Union[MsgBase, _Mapping]] = ..., shardName: _Optional[str] = ...) -> None: ...

class KeyDataPair(_message.Message):
    __slots__ = ["data", "key"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    key: str
    def __init__(self, key: _Optional[str] = ..., data: _Optional[bytes] = ...) -> None: ...

class KeyValuePair(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class MsgBase(_message.Message):
    __slots__ = ["msgID", "msg_type", "sourceID", "targetID", "timestamp"]
    MSGID_FIELD_NUMBER: _ClassVar[int]
    MSG_TYPE_FIELD_NUMBER: _ClassVar[int]
    SOURCEID_FIELD_NUMBER: _ClassVar[int]
    TARGETID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    msgID: int
    msg_type: MsgType
    sourceID: int
    targetID: int
    timestamp: int
    def __init__(self, msg_type: _Optional[_Union[MsgType, str]] = ..., msgID: _Optional[int] = ..., timestamp: _Optional[int] = ..., sourceID: _Optional[int] = ..., targetID: _Optional[int] = ...) -> None: ...

class MsgHeader(_message.Message):
    __slots__ = ["base"]
    BASE_FIELD_NUMBER: _ClassVar[int]
    base: MsgBase
    def __init__(self, base: _Optional[_Union[MsgBase, _Mapping]] = ...) -> None: ...

class PlaceholderGroup(_message.Message):
    __slots__ = ["placeholders"]
    PLACEHOLDERS_FIELD_NUMBER: _ClassVar[int]
    placeholders: _containers.RepeatedCompositeFieldContainer[PlaceholderValue]
    def __init__(self, placeholders: _Optional[_Iterable[_Union[PlaceholderValue, _Mapping]]] = ...) -> None: ...

class PlaceholderValue(_message.Message):
    __slots__ = ["tag", "type", "values"]
    TAG_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    tag: str
    type: PlaceholderType
    values: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, tag: _Optional[str] = ..., type: _Optional[_Union[PlaceholderType, str]] = ..., values: _Optional[_Iterable[bytes]] = ...) -> None: ...

class PrivilegeExt(_message.Message):
    __slots__ = ["object_name_index", "object_name_indexs", "object_privilege", "object_type"]
    OBJECT_NAME_INDEXS_FIELD_NUMBER: _ClassVar[int]
    OBJECT_NAME_INDEX_FIELD_NUMBER: _ClassVar[int]
    OBJECT_PRIVILEGE_FIELD_NUMBER: _ClassVar[int]
    OBJECT_TYPE_FIELD_NUMBER: _ClassVar[int]
    object_name_index: int
    object_name_indexs: int
    object_privilege: ObjectPrivilege
    object_type: ObjectType
    def __init__(self, object_type: _Optional[_Union[ObjectType, str]] = ..., object_privilege: _Optional[_Union[ObjectPrivilege, str]] = ..., object_name_index: _Optional[int] = ..., object_name_indexs: _Optional[int] = ...) -> None: ...

class Status(_message.Message):
    __slots__ = ["error_code", "reason"]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    error_code: ErrorCode
    reason: str
    def __init__(self, error_code: _Optional[_Union[ErrorCode, str]] = ..., reason: _Optional[str] = ...) -> None: ...

class ErrorCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class IndexState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class SegmentState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class PlaceholderType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MsgType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DslType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class CompactionState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ConsistencyLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ImportState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ObjectType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ObjectPrivilege(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class StateCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class LoadState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import time

import numpy as np
from pymilvus import (
    connections,
    utility,
    FieldSchema, CollectionSchema, DataType,
    Collection,
)

import threading
import concurrent

fmt = "\n=== {:30} ==="
search_latency_fmt = "search latency = {:.4f}s"
num_entities, dim = 3000, 8

print(fmt.format("start connecting to Milvus"))
connections.connect("default", host="localhost", port="19530")

if utility.has_collection("hello_milvus"):
    utility.drop_collection("hello_milvus")
    print(f"Dropping existing collection hello_milvus")

fields = [
    FieldSchema(name="pk", dtype=DataType.INT64, is_primary=True, auto_id=False),
    FieldSchema(name="random", dtype=DataType.DOUBLE),
    FieldSchema(name="embeddings", dtype=DataType.FLOAT_VECTOR, dim=dim)
]

schema = CollectionSchema(fields, "hello_milvus is the simplest demo to introduce the APIs")

print(fmt.format("Create collection `hello_milvus`"))
hello_milvus = Collection("hello_milvus", schema, consistency_level="Strong")


class MilvusMultiThreadingInsert:
    def __init__(self, collection_name: str, number_of_batch: int):

        self.thread_local = threading.local()
        self.collection_name = collection_name
        self.batchs = [i for i in range(number_of_batch)]

    def get_thread_local_collection(self):
        if not hasattr(self.thread_local, "collection"):
            self.thread_local.collection = Collection(self.collection_name)
        return self.thread_local.collection

    def insert_data(self, number: int):
        print(fmt.format(f"No.{number:2}: Start inserting entities"))
        rng = np.random.default_rng(seed=number)
        entities = [
            [i for i in range(num_entities)],
            rng.random(num_entities).tolist(),
            rng.random((num_entities, dim)),
        ]

        insert_result = hello_milvus.insert(entities)
        assert len(insert_result.primary_keys) == num_entities
        print(fmt.format(f"No.{number:2}: Finish inserting entities"))

    def insert_all_batches(self):
        with concurrent.futures.ThreadPoolExecutor(max_workers=12) as executor:
            executor.map(self.insert_data, self.batchs)

    def run(self):
        start_time = time.time()
        self.insert_all_batches()
        duration = time.time() - start_time
        print(f'Inserted {len(self.batchs)} batches of {num_entities} entities in {duration} seconds')
        print(f"Expected num_entities: {len(self.batchs)*num_entities}. \
                Acutal num_entites: {self.get_thread_local_collection().num_entities}")


if __name__ == "__main__":
    multithreading_insert = MilvusMultiThreadingInsert("hello_milvus", 10)
    multithreading_insert.run()



# Copyright (C) 2019-2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

from .client.stub import Milvus
from .client.prepare import Prepare
from .client.types import (
    Status,
    DataType,
    RangeType,
    IndexType,
    Replica,
    Group,
    Shard,
    BulkInsertState,
    ResourceGroupInfo
)
from .exceptions import (
    ParamError,
    MilvusException,
    MilvusUnavailableException,
    ExceptionsMessage
)
from .client import __version__

from .settings import DEBUG_LOG_LEVEL, INFO_LOG_LEVEL, WARN_LOG_LEVEL, ERROR_LOG_LEVEL

from .orm.collection import Collection
from .orm.connections import connections, Connections

from .orm.index import Index
from .orm.partition import Partition
from .orm.utility import (
    loading_progress,
    index_building_progress,
    wait_for_loading_complete,
    wait_for_index_building_complete,
    has_collection,
    has_partition,
    list_collections,
    drop_collection,
    get_query_segment_info,
    load_balance,
    mkts_from_hybridts, mkts_from_unixtime, mkts_from_datetime,
    hybridts_to_unixtime, hybridts_to_datetime,
    do_bulk_insert, get_bulk_insert_state, list_bulk_insert_tasks,
    reset_password, create_user, update_password, delete_user, list_usernames,
    create_resource_group, drop_resource_group, describe_resource_group,
    list_resource_groups, transfer_node, transfer_replica
)

from .orm import utility
from .orm.default_config import DefaultConfig, ENV_CONNECTION_CONF, DEFAULT_RESOURCE_GROUP

from .orm.search import SearchResult, Hits, Hit
from .orm.schema import FieldSchema, CollectionSchema
from .orm.future import SearchFuture, MutationFuture
from .orm.role import Role

__all__ = [
    'Collection', 'Index', 'Partition',
    'connections',
    'loading_progress', 'index_building_progress', 'wait_for_loading_complete', 'has_collection', 'has_partition',
    'list_collections', 'wait_for_loading_complete', 'wait_for_index_building_complete', 'drop_collection',
    'mkts_from_hybridts', 'mkts_from_unixtime', 'mkts_from_datetime',
    'hybridts_to_unixtime', 'hybridts_to_datetime',
    'reset_password', 'create_user', 'update_password', 'delete_user', 'list_usernames',
    'SearchResult', 'Hits', 'Hit', 'Replica', 'Group', 'Shard',
    'FieldSchema', 'CollectionSchema',
    'SearchFuture', 'MutationFuture',
    'utility', 'DefaultConfig', 'ExceptionsMessage', 'MilvusUnavailableException', 'BulkInsertState',
    'Role',
    'create_resource_group', 'drop_resource_group', 'describe_resource_group',
    'list_resource_groups', 'transfer_node', 'transfer_replica',

    'Milvus', 'Prepare', 'Status', 'DataType',
    'MilvusException',
    '__version__'
]



import sys
import datetime
from typing import Any, Union
from ..exceptions import ParamError
from ..grpc_gen import milvus_pb2 as milvus_types
from .singleton_utils import Singleton
from .utils import (
    valid_index_types,
    valid_binary_index_types,
    valid_index_params_keys,
)


def is_legal_address(addr: Any) -> bool:
    if not isinstance(addr, str):
        return False

    a = addr.split(":")
    if len(a) != 2:
        return False

    if not is_legal_host(a[0]) or not is_legal_port(a[1]):
        return False

    return True


def is_legal_host(host: Any) -> bool:
    if not isinstance(host, str):
        return False

    if len(host) == 0:
        return False

    return True


def is_legal_port(port: Any) -> bool:
    if isinstance(port, (str, int)):
        try:
            int(port)
        except ValueError:
            return False
        else:
            return True
    return False


def is_legal_vector(array: Any) -> bool:
    if not array or \
            not isinstance(array, list) or \
            len(array) == 0:
        return False

    # for v in array:
    #     if not isinstance(v, float):
    #         return False

    return True


def is_legal_bin_vector(array: Any) -> bool:
    if not array or \
            not isinstance(array, bytes) or \
            len(array) == 0:
        return False

    return True


def is_legal_numpy_array(array: Any) -> bool:
    return not (array is None or array.size == 0)


def int_or_str(item: Union[int, str]) -> str:
    if isinstance(item, int):
        return str(item)

    return item


def is_correct_date_str(param: str) -> bool:
    try:
        datetime.datetime.strptime(param, '%Y-%m-%d')
    except ValueError:
        return False

    return True


def is_legal_dimension(dim: Any) -> bool:
    return isinstance(dim, int)


def is_legal_index_size(index_size: Any) -> bool:
    return isinstance(index_size, int)


def is_legal_table_name(table_name: Any) -> bool:
    return table_name and isinstance(table_name, str)


def is_legal_field_name(field_name: Any) -> bool:
    return field_name and isinstance(field_name, str)


def is_legal_nlist(nlist: Any) -> bool:
    return not isinstance(nlist, bool) and isinstance(nlist, int)


def is_legal_topk(topk: Any) -> bool:
    return not isinstance(topk, bool) and isinstance(topk, int)


def is_legal_ids(ids: Any) -> bool:
    if not ids or not isinstance(ids, list):
        return False

    # TODO: Here check id valid value range may not match other SDK
    for i in ids:
        if not isinstance(i, (int, str)):
            return False
        try:
            i_ = int(i)
            if i_ < 0 or i_ > sys.maxsize:
                return False
        except Exception:
            return False

    return True


def is_legal_nprobe(nprobe: Any) -> bool:
    return isinstance(nprobe, int)


def is_legal_cmd(cmd: Any) -> bool:
    return cmd and isinstance(cmd, str)


def parser_range_date(date: Union[str, datetime.date]) -> str:
    if isinstance(date, datetime.date):
        return date.strftime('%Y-%m-%d')

    if isinstance(date, str):
        if not is_correct_date_str(date):
            raise ParamError(message='Date string should be YY-MM-DD format!')

        return date

    raise ParamError(message='Date should be YY-MM-DD format string or datetime.date, '
                     'or datetime.datetime object')


def is_legal_date_range(start: str, end: str) -> bool:
    start_date = datetime.datetime.strptime(start, "%Y-%m-%d")
    end_date = datetime.datetime.strptime(end, "%Y-%m-%d")
    if (end_date - start_date).days < 0:
        return False

    return True


def is_legal_partition_name(tag: Any) -> bool:
    return tag is not None and isinstance(tag, str)


def is_legal_limit(limit: Any) -> bool:
    return isinstance(limit, int) and limit > 0


def is_legal_anns_field(field: Any) -> bool:
    return field and isinstance(field, str)


def is_legal_search_data(data: Any) -> bool:
    import numpy as np
    if not isinstance(data, (list, np.ndarray)):
        return False

    for vector in data:
        # list -> float vector
        # bytes -> byte vector
        if not isinstance(vector, (list, bytes, np.ndarray)):
            return False

    return True


def is_legal_output_fields(output_fields: Any) -> bool:
    if output_fields is None:
        return True

    if not isinstance(output_fields, list):
        return False

    for field in output_fields:
        if not is_legal_field_name(field):
            return False

    return True


def is_legal_partition_name_array(tag_array: Any) -> bool:
    if tag_array is None:
        return True

    if not isinstance(tag_array, list):
        return False

    for tag in tag_array:
        if not is_legal_partition_name(tag):
            return False

    return True

def is_legal_replica_number(replica_number: int) -> bool:
    return isinstance(replica_number, int)

# https://milvus.io/cn/docs/v1.0.0/metric.md#floating
def is_legal_index_metric_type(index_type: str, metric_type: str) -> bool:
    if index_type not in ("FLAT",
                          "IVF_FLAT",
                          "IVF_SQ8",
                          # "IVF_SQ8_HYBRID",
                          "IVF_PQ",
                          "HNSW",
                          # "NSG",
                          "ANNOY",
                          "RHNSW_FLAT",
                          "RHNSW_PQ",
                          "RHNSW_SQ",
                          "AUTOINDEX",
                          "DISKANN"):
        return False
    if metric_type not in ("L2", "IP"):
        return False
    return True


# https://milvus.io/cn/docs/v1.0.0/metric.md#binary
def is_legal_binary_index_metric_type(index_type: str, metric_type: str) -> bool:
    if index_type == "BIN_FLAT":
        if metric_type in ("JACCARD", "TANIMOTO", "HAMMING", "SUBSTRUCTURE", "SUPERSTRUCTURE"):
            return True
    elif index_type == "BIN_IVF_FLAT":
        if metric_type in ("JACCARD", "TANIMOTO", "HAMMING"):
            return True
    return False


def _raise_param_error(param_name: str, param_value: Any) -> None:
    raise ParamError(message=f"`{param_name}` value {param_value} is illegal")


def is_legal_round_decimal(round_decimal: Any) -> bool:
    return isinstance(round_decimal, int) and -2 < round_decimal < 7


def is_legal_travel_timestamp(ts: Any) -> bool:
    return isinstance(ts, int) and ts >= 0


def is_legal_guarantee_timestamp(ts: Any) -> bool:
    return isinstance(ts, int) and ts >= 0


def is_legal_user(user) -> bool:
    return isinstance(user, str)


def is_legal_password(password) -> bool:
    return isinstance(password, str)


def is_legal_role_name(role_name: Any) -> bool:
    return role_name and isinstance(role_name, str)


def is_legal_operate_user_role_type(operate_user_role_type: Any) -> bool:
    return operate_user_role_type in \
        (milvus_types.OperateUserRoleType.AddUserToRole, milvus_types.OperateUserRoleType.RemoveUserFromRole)


def is_legal_include_user_info(include_user_info: Any) -> bool:
    return isinstance(include_user_info, bool)


def is_legal_include_role_info(include_role_info: Any) -> bool:
    return isinstance(include_role_info, bool)


def is_legal_object(object: Any) -> bool:
    return object and isinstance(object, str)


def is_legal_object_name(object_name: Any) -> bool:
    return object_name and isinstance(object_name, str)


def is_legal_privilege(privilege: Any) -> bool:
    return privilege and isinstance(privilege, str)


def is_legal_collection_properties(properties: Any) -> bool:
    return properties and isinstance(properties, dict)


def is_legal_operate_privilege_type(operate_privilege_type: Any) -> bool:
    return operate_privilege_type in \
            (milvus_types.OperatePrivilegeType.Grant, milvus_types.OperatePrivilegeType.Revoke)


class ParamChecker(metaclass=Singleton):
    def __init__(self) -> None:
        self.check_dict = {
            "collection_name": is_legal_table_name,
            "field_name": is_legal_field_name,
            "dimension": is_legal_dimension,
            "index_file_size": is_legal_index_size,
            "topk": is_legal_topk,
            "ids": is_legal_ids,
            "nprobe": is_legal_nprobe,
            "nlist": is_legal_nlist,
            "cmd": is_legal_cmd,
            "partition_name": is_legal_partition_name,
            "partition_name_array": is_legal_partition_name_array,
            "limit": is_legal_limit,
            "anns_field": is_legal_anns_field,
            "search_data": is_legal_search_data,
            "output_fields": is_legal_output_fields,
            "round_decimal": is_legal_round_decimal,
            "travel_timestamp": is_legal_travel_timestamp,
            "guarantee_timestamp": is_legal_guarantee_timestamp,
            "user": is_legal_user,
            "password": is_legal_password,
            "role_name": is_legal_role_name,
            "operate_user_role_type": is_legal_operate_user_role_type,
            "include_user_info": is_legal_include_user_info,
            "include_role_info": is_legal_include_role_info,
            "object": is_legal_object,
            "object_name": is_legal_object_name,
            "privilege": is_legal_privilege,
            "operate_privilege_type": is_legal_operate_privilege_type,
            "properties": is_legal_collection_properties,
            "replica_number": is_legal_replica_number,
            "resource_group_name": is_legal_table_name,
        }

    def check(self, key, value):
        if key in self.check_dict:
            if not self.check_dict[key](value):
                _raise_param_error(key, value)
        else:
            raise ParamError(message=f"unknown param `{key}`")

def _get_param_checker():
    return ParamChecker()

def check_pass_param(*_args: Any, **kwargs: Any) -> None:  # pylint: disable=too-many-statements
    if kwargs is None:
        raise ParamError(message="Param should not be None")
    checker = _get_param_checker()
    for key, value in kwargs.items():
        checker.check(key, value)


def check_index_params(params):
    params = params or {}
    if not isinstance(params, dict):
        raise ParamError(message="Params must be a dictionary type")
    # params preliminary validate
    if 'index_type' not in params:
        raise ParamError(message="Params must contains key: 'index_type'")
    if 'params' not in params:
        raise ParamError(message="Params must contains key: 'params'")
    if 'metric_type' not in params:
        raise ParamError(message="Params must contains key: 'metric_type'")
    if not isinstance(params['params'], dict):
        raise ParamError(message="Params['params'] must be a dictionary type")
    if params['index_type'] not in valid_index_types:
        raise ParamError(message=f"Invalid index_type: {params['index_type']}, which must be one of: {str(valid_index_types)}")
    for k in params['params'].keys():
        if k not in valid_index_params_keys:
            raise ParamError(message=f"Invalid params['params'].key: {k}")
    for v in params['params'].values():
        if not isinstance(v, int):
            raise ParamError(message=f"Invalid params['params'].value: {v}, which must be an integer")
    # filter invalid metric type
    if params['index_type'] in valid_binary_index_types:
        if not is_legal_binary_index_metric_type(params['index_type'], params['metric_type']):
            raise ParamError(message=f"Invalid metric_type: {params['metric_type']}, which does not match the index type: {params['index_type']}")
    else:
        if not is_legal_index_metric_type(params['index_type'], params['metric_type']):
            raise ParamError(message=f"Invalid metric_type: {params['metric_type']}, which does not match the index type: {params['index_type']}")



import pathlib
import setuptools

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setuptools.setup(
    name="pymilvus",
    author='Milvus Team',
    author_email='milvus-team@zilliz.com',
    setup_requires=['setuptools_scm'],
    use_scm_version={'local_scheme': 'no-local-version', 'version_scheme': 'release-branch-semver'},
    description="Python Sdk for Milvus",
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/milvus-io/pymilvus',
    license="Apache-2.0",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "grpcio>=1.49.1,<=1.53.0",
        "protobuf>=3.20.0",
        "mmh3>=2.0",
        "ujson>=2.0.0",
        "pandas>=1.2.4",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],

    python_requires='>=3.7'
)
